﻿#include "mainwindow.h"
#include "ui_mainwindow.h"
// initialize config file
QSettings settings("config.ini", QSettings::IniFormat);
// initialize a vector for storing the clients' data
QVector<dataClass *> dataList(1000000);

MainWindow::MainWindow(QWidget *parent)

    : QMainWindow(parent),
      ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    QString year = QDateTime::currentDateTime().toString("yyyy");
    // check if the certificates are present and generate new ones if not
    if (!QFile::exists("cert"+year+".pem") || !QFile::exists("key"+year+".pem")){
        QMessageBox::critical(0, "Certificates not found", "The certificates for secure TLS connection are not found. They will be generated now. This should take less than a minute.");
        QProcess openssl;
        QStringList args;
        args << "/C" << QCoreApplication::applicationDirPath()+"/openssl.exe" << "req" << "-new" << "-config" << QCoreApplication::applicationDirPath()+"/openssl.cfg" <<
                "-newkey" << "rsa:4096" << "-days" << "365" << "-nodes" << "-x509" << "-subj" << "/C=BG/ST=Blagoevgrad/L=Blagoevgrad/O=AUBG/CN=CheatingPreventionSoftware" <<
                "-keyout" << "key"+year+".pem" << "-out" << "cert"+year+".pem";
        if (QProcess::execute("C:/windows/system32/cmd.exe", args) < 0){
            QMessageBox::critical(0, "Error generating certificates", "The certificates cannot be generated. Check if the openssl.exe file is present and the path does not contain spaces.");
            std::exit(EXIT_FAILURE);
        } else {
            if (!QFile::exists("cert"+year+".pem") || !QFile::exists("key"+year+".pem")){
                QMessageBox::critical(0, "Error generating certificates", "The certificates cannot be generated. Check if the openssl.exe file is present and the path does not contain spaces.");
                std::exit(EXIT_FAILURE);
            }
        }
    }
    // initialize logging function
    QString path = QCoreApplication::applicationDirPath() + "/logs/";
    QDir directory;
    if(!directory.exists(path))
        directory.mkpath(path);
    logFile = new QFile(path + QDateTime::currentDateTime().toString("yyyy.MM.dd_hh.mm") + ".html");
    logFile->open(QIODevice::WriteOnly);
    // initialize the settings from the config file
    QString port = settings.value("Port", "23000").toString();
    forbiddenApps = settings.value("Forbidden_Applications", QVariant(QStringList({"TeamViewer.exe", "AnyDesk.exe", "LogMeIn.exe", "rserver3.exe", "ZA_Access.exe", "SRService.exe",
                                                                                   "RPCSuite.exe"}))).value<QStringList>();
    keywords = settings.value("Suspicious_Keywords", QVariant(QStringList({"Search", "Quizlet", "Brainly", "Chegg", "Word", "PowerPoint", "Messenger"}))).value<QStringList>();
    settings.setValue("Forbidden_Applications", forbiddenApps);
    settings.setValue("Suspicious_Keywords", keywords);
    // prompt the user to choose a port, showing the last selected port as default and checking if the port number is valid
    bool occupiedPort;
    do{
        bool ok;
        port = QInputDialog::getText(0, "Port Selection", "Choose the port number between 1 and 65535 (default 23000):", QLineEdit::Normal, port, &ok);
        if (!ok)
            std::exit(EXIT_FAILURE);
        settings.setValue("Port", port);
        if ((port.toInt()>=1 && port.toInt()<=65535)){
            if(checkPort(port.toInt())){
                QMessageBox::critical(0, "Port Conflict", "The selected port is already occupied, choose another port!");
                occupiedPort = true;
            } else
                occupiedPort = false;
        }
    } while (occupiedPort || port.isEmpty() || !(port.toInt()>=1 && port.toInt()<=65535));
    // invoke a function that displays computer's IP addresses
    displayIPaddresses(port);
    // initialize and start the networking thread
    networkThread = new Networking(this, port.toInt());
    networkThread->start();
    // link the update signal of the networking class to the updater function of the mainwindow class
    connect(networkThread, SIGNAL(update(QString)), this, SLOT(updater(QString)));
    // intialize and start the timer that invokes checkConnection function every 3 seconds
    QTimer *timer = new QTimer;
    connect(timer,SIGNAL(timeout()), this, SLOT(checkConnection()));
    timer->start(3000);
}

MainWindow::~MainWindow()
{
    // terminate the networking thread and close the log file before exiting the application
    networkThread->terminate();
    logFile->close();
    delete ui;
}

void MainWindow::updater(QString string){
    // check if the data is correctly formatted, coming from the appropriate client application
    // which always sends six digits of the ID number followed by a semicolon (;)
    bool ok;
    int id = string.left(6).toInt(&ok);
    if (ok && string.at(6)==';'){
        // check if the ID is new, that is the client is sending data for the first time
        if (dataList.at(id)==nullptr){
            ui->label->hide(); // hide the "Waiting for connections..." sign
            dataList[id] = new dataClass(string); // initialize a dataclass object for storing data
            // initialize a label widget for displaying data
            dataList.at(id)->label = new QLabel(this);
            dataList.at(id)->label->setWordWrap(true);
            dataList.at(id)->label->setStyleSheet("border: 1px solid black; background-color: #5BC85B;");
            // replace the text in the label with the result of the getResult function
            dataList.at(id)->label->setText(dataList.at(id)->getResult(keywords, forbiddenApps));
            // calculate using the number of clients from the count variable the row and the column numbers to place the label widgets in a grid
            int row=0;
            int column=0;
            count++;
            int side = (int)ceil(sqrt(count) - 0.000001) - 1;
            if (count <= side * (side + 1)) {
            row = count - 1 - side * side;
            column = side;
            } else {
            row = side;
            column = count - 1 - side * (side + 1);
            }
            // add the new label to the corresponding place in the grid layout
            ui->gridLayout->addWidget(dataList.at(id)->label,row,column);
            // link the linkActivated signal of the label to the addKeywordsWindow function of the mainwindow class
            connect(dataList.at(id)->label, SIGNAL(linkActivated(QString)), this,
                    SLOT(addKeywordsWindow(QString)));

        } else {
            // if the client is already initialized and sends the data for the second time,
            // update the data inside the dataclass instance and display the result of getResult in the corresponding label
            dataList.at(id)->updateData(string);
            dataList.at(id)->label->setText(dataList.at(id)->getResult(keywords, forbiddenApps));
        }
    }
}

void MainWindow::checkConnection()
{
    // this function provides checking if the client is connected by comparing the previosly checked timestamp
    // with the last timestamp which should be greater
    QTextStream out(logFile);
    for(int i=0;i<1000000;i++){
        if (dataList.at(i)!=nullptr){
            if (dataList.at(i)->time > dataList.at(i)->previousTime){
                dataList.at(i)->previousTime=dataList.at(i)->time;
                dataList.at(i)->label->setStyleSheet("border: 1px solid black; background-color: #5BC85B;");
            } else {
                dataList.at(i)->label->setStyleSheet("border: 1px solid black; background-color: #EE9090;");
            }
            // at the same time, we check if the data should be added to the log by using the makeLogEntry boolean variable
            if(dataList.at(i)->makeLogEntry){
                // add the timestamp to the log file followed by the data containing suspicious behavior
                out << "<h1>" + QDateTime::currentDateTime().toString("hh:mm:ss") + "</h1>";
                out << dataList.at(i)->getResult(keywords, forbiddenApps) << "<p>";
                dataList.at(i)->makeLogEntry=false;
            }
        }
    }
}

void MainWindow::addKeywordsWindow(QString title)
{
    // create a new window that is not resizeable
    keywordsWindow = new QWidget;
    keywordsWindow->setWindowFlags(Qt::Dialog);
    // take all words from the window title
    QStringList splittedTitle = title.split(" ");
    // display those words vertically as buttons
    QVBoxLayout *verticalLayout = new QVBoxLayout;
    keywordsWindow->setLayout(verticalLayout);
    for (int i=0;i<splittedTitle.length();i++){
        QPushButton *btn = new QPushButton();
        btn->setText(splittedTitle.at(i));
        for(int j=0;j<keywords.length();j++){
            if (splittedTitle.at(i) == keywords.at(j)){ // if the word is already on the list, highlight it in red
                btn->setStyleSheet("color: red;");
            }
        }
        // link the clicking of the button to the modifyKeywords function
        connect(btn,SIGNAL(clicked()), this, SLOT(modifyKeywords()));
        verticalLayout->addWidget(btn);
    }
    keywordsWindow->show();
}

void MainWindow::modifyKeywords()
{
    // get the clicked word from the button that was clicked
    QString keyword = qobject_cast<QPushButton*>(sender())->text();
    // check if the keyword is already in the suspicious keywords list
    int index = keywords.lastIndexOf(keyword);
    if (index == -1){
        keywords.append(keyword); // add a new keyword
    } else {
        keywords.removeAt(index); // remove an existing keyword
    }
    settings.setValue("Suspicious_Keywords", keywords); // save the modified list to the config file
    keywordsWindow->deleteLater(); // close the keywords modification window
}

bool MainWindow::checkPort(int PORT)
{
    // the function attempts connecting to the localhost at the selected port
    // returns false if the port is free, returns true if the port is occupied
    WSADATA winsock;
    WSAStartup(514, &winsock);
    SOCKET clientSocket = socket(AF_INET, SOCK_STREAM, 0);
    struct sockaddr_in init;
    init.sin_port = htons(PORT);
    init.sin_family = AF_INET;
    init.sin_addr.s_addr = inet_addr("127.0.0.1");
    if(WSAAPI::connect(clientSocket, (sockaddr*)&init, sizeof(init))==SOCKET_ERROR){
        closesocket(clientSocket);
        WSACleanup();
        return false;
    } else {
        closesocket(clientSocket);
        WSACleanup();
        return true;
    }
}

void MainWindow::displayIPaddresses(QString port){
    // get the computer's global IP address by reading the contents of http://api.ipify.org/ website
    DWORD size;
    char ipAddress[20];
    int tries = 0;
    do{
    HINTERNET internet = InternetOpen(NULL,INTERNET_OPEN_TYPE_PRECONFIG,NULL,NULL,0);
    HINTERNET file = InternetOpenUrlA(internet,"http://api.ipify.org/",NULL,0,INTERNET_FLAG_RELOAD,0);
    InternetReadFile(file,&ipAddress,sizeof(ipAddress),&size);
    InternetCloseHandle(file);
    InternetCloseHandle(internet);
    tries++;
    } while (size==0 && tries<=3);
    // add the IP address to a QString variable and append the port number if it is not the default port
    QString output;
    output.append(ipAddress);
    if (output.length()>1){
        while(!output.at(output.length()-1).isDigit())
            output.chop(1);
        if(output.at(output.length()-1).isDigit()){ // check if the response contains an IP address
            if (port!="23000")
                output.append(":"+port);
            output.append("\n");
        } else
            QMessageBox::critical(0, "Error Getting Global IP", "Unable to detect your global IP address");
    } else
        QMessageBox::critical(0, "Error Getting Global IP", "Unable to detect your global IP address");
    // get the computer's local IP addresses
    WSAData wsadata;
    WSAStartup(MAKEWORD(2, 2), &wsadata);
    char hostname[100];
    gethostname(hostname, 100);
    struct hostent *host = gethostbyname(hostname);
    for (int i=0;host->h_addr_list[i]!=0;++i){
        struct in_addr ip;
        memcpy(&ip, host->h_addr_list[i], 4);
        output.append(inet_ntoa(ip));
        // add the local IP addresses to the QString variable and append the port number if it is not the default port
        if (port!="23000")
            output.append(":"+port);
        output.append("\n");
    }
    WSACleanup();
    output.chop(1);
    // show a GUI window with the IP addresses
    QInputDialog::getMultiLineText(0, "Your IP Address", "Copy your IP address:", output);
}
